#! /usr/bin/env python

"""
Deal with GC3Pie configuration files.
"""

# Copyright (C) 2012-2016, 2018, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import absolute_import, print_function, unicode_literals

from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str

# stdlib imports
import inspect
import os
import re
import sys
if sys.version_info[0] == 2:
    from ConfigParser import SafeConfigParser
    from ConfigParser import Error as ConfigParserError
    def make_config_parser():
        return SafeConfigParser()
    def read_config_lines(parser, stream, source):
        return parser.readfp(stream, source)
else:
    # `SafeConfigParser` was deprecated in Py3 in favor of `ConfigParser`
    from configparser import ConfigParser
    from configparser import Error as ConfigParserError
    def make_config_parser():
        return ConfigParser(strict=False)
    # `readfp()` is deprecated since Py3.2
    def read_config_lines(parser, stream, source):
        return parser.read_file(stream, source)

# GC3Pie imports
import gc3libs
import gc3libs.defaults
import gc3libs.authentication
from collections import defaultdict
import gc3libs.utils

from gc3libs.quantity import Memory, GB, MB, MiB, Duration, hours


# module metadata
__docformat__ = 'reStructuredText'


# auxiliary methods for `Configuration`
#
# these must be defined before `Configuration` is parsed, because they
# are referenced in the definition of `Configuration` itself
#

# map values for the `architecture=...` configuration item
# into internal constants
_architecture_value_map = {
    # 'x86-32', 'x86 32-bit', '32-bit x86' and variants thereof
    re.compile('x86[ _-]+32([ _-]*bits?)?', re.I): gc3libs.Run.Arch.X86_32,
    re.compile('32[ _-]*bits? +[ix]86', re.I): gc3libs.Run.Arch.X86_32,
    # accept also values printed by `uname -a` on 32-bit x86 archs
    re.compile('i[3456]86', re.I): gc3libs.Run.Arch.X86_32,
    # 'x86_64', 'x86 64-bit', '64-bit x86' and variants thereof
    re.compile('x86[ _-]+64([ _-]*bits?)?', re.I): gc3libs.Run.Arch.X86_64,
    re.compile('64[ _-]*bits? +[ix]86', re.I): gc3libs.Run.Arch.X86_64,
    # also accept commercial arch names
    re.compile('(amd[ -]*64|x64|emt64|intel[ -]*64)( *bits?)?', re.I):
    gc3libs.Run.Arch.X86_64,
    # finally, map "32-bit" and "64-bit" to i686 and x86_64
    re.compile('32[ _-]*bits?', re.I): gc3libs.Run.Arch.X86_32,
    re.compile('64[ _-]*bits?', re.I): gc3libs.Run.Arch.X86_64,
}

def _matching_architecture(value):
    """Return first matching entry from `_architecture_value_map`."""
    for matcher, arch in _architecture_value_map.items():
        if matcher.match(value):
            return arch
    raise ValueError("Unknown architecture '%s'." % value)

def _parse_architecture(arch_str):
    """Return set of values in `Run.Arch` corresponding to a comma-separated list of architecture strings."""
    archs = [_matching_architecture(value.strip()) for value in arch_str.split(',')]
    if len(archs) == 0:
        raise ValueError("Empty or invalid 'architecture' setting.")
    return set(archs)


def _legacy_parse_duration(duration_str):
    try:
        # old-style config: integral number of hours
        val = int(duration_str) * hours
        gc3libs.log.warning("'max_walltime' should always have a "
                            "valid unit format (e.g. '24 hours'). Using "
                            "default unit: hours")
        return val
    except ValueError:
        # apply `Duration` parsing rules; if this fails, users will
        # see the error message from the `Duration` parser.
        return Duration(duration_str)


def _legacy_parse_memory(memory_str, name, unit=GB, unit_name='GB'):
    try:
        # old-style config: integral number of "unit"
        val = int(memory_str) * unit
        gc3libs.log.warning(
            "'%s' should always have a valid unit format (e.g. '2 GB')."
            " Using default unit: %s", name, unit_name)
        return val
    except ValueError:
        # apply usual quantity parsing rules; if this fails, users
        # will see the error message from the `Memory`/`Quantity` parser.
        return Memory(memory_str)


def _legacy_parse_os_overhead(os_overhead_str):
    try:
        # old-style config: integral number of MiBs
        val = int(os_overhead_str) * MiB
        gc3libs.log.warning("'vm_os_overhead' should always have a "
                            "valid unit format (e.g. '512 MiB'). Using "
                            "default unit: MiB")
        return val
    except ValueError:
        # apply usual quantity parsing rules; if this fails, users
        # will see the error message from the `Memory`/`Quantity` parser.
        return Memory(os_overhead_str)

# the main class of this module


class Configuration(gc3libs.utils.Struct):

    """
    In-memory representation of the GC3Pie configuration.

    This class provides facilities for:

    * parsing configuration files (methods `load`:meth: and
      `merge_file`:meth:);
    * validating the loaded values;
    * instanciating the internal GC3Pie objects resulting from the
      configuration (methods `make_auth`:meth: and
      `make_resource`:meth:).

    The constructor takes a list of files to load (`locations`) and a
    list of key=value pairs to provide defaults for the configuration.
    Both lists are optional and can be omitted, resulting in a
    configuration containing only GC3Pie default values.

    Example 1: initialization from config file::

      >>> import os
      >>> example_cfgfile = os.path.join(
      ...    os.path.dirname(__file__), 'etc/gc3pie.conf.example')
      >>> cfg = Configuration(example_cfgfile)
      >>> cfg.debug
      '0'

    Example 2: initialization from key=value list::

      >>> cfg = Configuration(auto_enable_auth=False, foo=1, bar='baz')
      >>> cfg.auto_enable_auth
      False
      >>> cfg.foo == 1
      True
      >>> cfg.bar == 'baz'
      True

    When both a configuration file *and* a key=value list is present,
    values in the configuration files override those in the key=value
    list::

      >>> cfg = Configuration(example_cfgfile, debug=1)
      >>> cfg.debug == '0'
      True

    Example 3: default initialization::

      >>> cfg = Configuration()
      >>> cfg.auto_enable_auth
      True

    """

    def __init__(self, *locations, **extra_args):
        self._auth_factory = None

        # these fields are required
        self.resources = defaultdict(gc3libs.utils.Struct)
        self.auths = defaultdict(gc3libs.utils.Struct)

        # ensure `auth = none` is always available
        self.auths['none'] = gc3libs.utils.Struct(type='none')

        # use keyword arguments to set defaults
        self.auto_enable_auth = extra_args.pop('auto_enable_auth', True)
        self.update(extra_args)

        # save the list of (valid) config files
        self.cfgfiles = []

        # load configuration files if any
        if locations:
            self.load(*locations)

        # actual resource constructor classes
        self._resource_constructors_cache = {}


    def load(self, *locations):
        """
        Merge settings from configuration files into this `Configuration`
        instance.

        Environment variables and `~` references are expanded in the
        location file names.

        If any of the specified files does not exist or cannot be read
        (for whatever reason), a message is logged but the error is
        ignored.  However, a `NoConfigurationFile` exception is raised
        if *none* of the specified locations could be read.

        :raise gc3libs.exceptions.NoConfigurationFile:
            if none of the specified files could be read.
        """
        files_successfully_read = 0
        files_successfully_parsed = 0

        for filename in locations:
            filename = os.path.expandvars(os.path.expanduser(filename))
            if os.path.exists(filename):
                if not os.access(filename, os.R_OK):
                    gc3libs.log.debug(
                        "Configuration.load(): File '%s' cannot be read,"
                        " ignoring.", filename)
                    continue  # with next `filename`
            else:
                gc3libs.log.debug(
                    "Configuration.load(): File '%s' does not exist,"
                    " ignoring.", filename)
                continue  # with next `filename`

            filename = os.path.abspath(filename)
            self.cfgfiles.append(filename)
            # since file passed the `R_OK` access test, we know it
            # *can* be read before actually doing it
            files_successfully_read += 1
            try:
                self.merge_file(filename)
                files_successfully_parsed += 1
            except gc3libs.exceptions.ConfigurationError:
                continue  # with next file

        if files_successfully_read == 0:
            raise gc3libs.exceptions.NoAccessibleConfigurationFile(
                "Could not read any configuration file; tried location '%s'."
                % "', '".join(locations))
        if files_successfully_parsed == 0:
            raise gc3libs.exceptions.NoValidConfigurationFile(
                "Could not parse any configuration file;"
                " tried location(s) '%s' but they all had errors."
                " (Which see in previous log messages.)"
                % "', '".join(locations))

    def merge_file(self, filename):
        """
        Read configuration files and merge the settings into this
        `Configuration` object.

        Contrary to `load`:meth: (which see), the file name is taken
        literally and an error is raised if the file cannot be read
        for whatever reason.

        Any parameter which is set in the configuration files
        ``[DEFAULT]`` section, and whose name does not start with
        underscore (``_``) defines an attribute in the current
        `Configuration`.

        .. warning::

          No type conversion is performed on values set this way - so
          they all end up being strings!

        :raise gc3libs.exceptions.ConfigurationError: if the
            configuration file does not exist, cannot be read, is
            corrupt or has wrong format.
        """
        gc3libs.log.debug(
            "Configuration.merge_file(): Reading file '%s' ...",
            filename)
        with open(filename, 'r') as stream:
            defaults, resources, auths = self._parse(stream, filename)
        for name, values in resources.items():
            self.resources[name].update(values)
        for name, values in auths.items():
            self.auths[name].update(values)
        for name, value in defaults.items():
            if not name.startswith('_'):
                self[name] = value

    def _parse(self, stream, filename=None):
        """
        Read configuration file and return a `(defaults, resources, auths)`
        triple.

        The members of the result triple are as follows:

        * `defaults`: a dictionary containing keys found in the
          ``[DEFAULTS]`` section of the configuration file (if any);

        * `resources`: a dictionary mapping resource names into a
          dictionary of key/value attributes contained in the
          configuration file under the ``[resource/name]`` heading;

        * `auths`: same for the ``[auth/name]`` sections.

        In addition, key renaming (for compatibility with previous
        versions) and type conversion is performed here, so that the
        returned dictionaries conform to a specified schema.  Type
        conversion are performed according to the value of the
        `_convert` attribute on this object; any attribute not
        mentioned in that table will have type ``str`` (i.e., it is
        left unchanged).
        """
        defaults = dict()
        resources = defaultdict(dict)
        auths = defaultdict(dict)

        parser = make_config_parser()
        try:
            read_config_lines(parser, stream, filename)
        except ConfigParserError as err:
            if filename is None:
                try:
                    filename = stream.name
                except AttributeError:
                    filename = repr(stream)
            raise gc3libs.exceptions.ConfigurationError(
                "Configuration file '%s' is unreadable or malformed: %s: %s"
                % (filename, err.__class__.__name__, err))

        # update `defaults` with the contents of the `[DEFAULTS]` section
        defaults.update(parser.defaults())

        for sectname in parser.sections():
            if sectname.startswith('auth/'):
                # handle auth section
                name = sectname.split('/', 1)[1]
                config_items = dict(parser.items(sectname))
                gc3libs.log.debug(
                    "Config._parse():"
                    " Read configuration stanza for auth '%s'",
                    name)

                # minimal sanity check
                for key in self._auth_required_keys:
                    if key not in config_items:
                        raise gc3libs.exceptions.ConfigurationError(
                            "Missing mandatory configuration key `{key}`"
                            " in section [{sectname}]"
                            " of the configuration file `{filename}`."
                            " This configuration file will be ignored."
                            .format(
                                key=key,
                                sectname=sectname,
                                filename=filename
                            ))

                auths[name].update(config_items)
                auths[name]['name'] = name

            elif sectname.startswith('resource/'):
                # handle resource section
                name = sectname.split('/', 1)[1]
                gc3libs.log.debug(
                    "Config._parse():"
                    " Read configuration stanza for resource '%s'." %
                    name)

                config_items = dict(parser.items(sectname))
                self._perform_key_renames(
                    config_items, self._renamed_keys, filename)
                self._perform_value_updates(
                    config_items, self._updated_values, filename)
                self._perform_filename_conversion(
                    config_items, self._path_key_regexp, filename)
                try:
                    self._perform_type_conversions(
                        config_items, self._convert, filename)
                except Exception as err:
                    raise gc3libs.exceptions.ConfigurationError(
                        "Incorrect entry for resource '%s' in configuration"
                        " file '%s': %s"
                        % (name, filename, str(err)))

                # minimal sanity checks
                for key in self._resource_required_keys:
                    if key not in config_items:
                        raise gc3libs.exceptions.ConfigurationError(
                            "Missing mandatory configuration key `{key}`"
                            " in section [{sectname}]"
                            " of the configuration file `{filename}`."
                            " This configuration file will be ignored."
                            .format(
                                key=key,
                                sectname=sectname,
                                filename=filename
                            ))

                for key in self._resource_keys_value_required:
                    if key in config_items and not config_items[key]:
                        raise gc3libs.exceptions.ConfigurationError(
                            "Found empty configuration key `{key}`"
                            " in section [{sectname}]"
                            .format(key=key, sectname=sectname))

                resources[name].update(config_items)
                resources[name]['name'] = name
                if __debug__:
                    gc3libs.log.debug(
                        "Config._parse(): Resource '%s' defined by: %s.", name,
                        ', '.join([
                            ("%s=%r" %
                             (k, v)) for k, v in sorted(
                                 resources[name].items())]))

            else:
                # Unhandled sectname
                gc3libs.log.warning(
                    "Config._parse(): unknown configuration section '%s'"
                    " -- ignoring!",
                    sectname)

        return (defaults, resources, auths)

    # config keys common to every kind of `[auth/*]` section;
    # if any of these is missing, the section is clearly invalid
    _auth_required_keys = (
        'type',
    )

    # config keys common to every kind of `[resource/*]` section;
    # if any of these is missing, the section is clearly invalid
    _resource_required_keys = (
        'architecture',
        'max_cores',
        'max_cores_per_job',
        'max_memory_per_core',
        'max_walltime',
        'type',
    )

    # config keys that are not allowed to have empty values
    _resource_keys_value_required = (
        'keypair_name',
    )

    _renamed_keys = {
        # old key name           new key name
        # ===================    ===================
        'ncores'               : 'max_cores',
        'sge_accounting_delay' : 'accounting_delay',
    }

    @staticmethod
    def _perform_key_renames(config_items, renames, filename):
        for oldkey, newkey in renames.items():
            if oldkey in config_items:
                gc3libs.log.warning(
                    "Configuration item '%s' was renamed to '%s',"
                    " please change occurrences of '%s' to '%s'"
                    " in configuration file '%s'.",
                    oldkey, newkey, oldkey, newkey, filename)
                if newkey in config_items:
                    # drop
                    gc3libs.log.error(
                        "Both old-style configuration item '%s' and new-style"
                        " '%s' detected in file '%s': ignoring old-style item"
                        " '%s=%s'.",
                        oldkey,
                        newkey,
                        filename,
                        config_items[oldkey])
                else:
                    config_items[newkey] = config_items[oldkey]
                del config_items[oldkey]

    _updated_values = {
        # key name  old value             new value
        # ========  ===================   ==================
        'type': {
                    'arc'               : gc3libs.defaults.ARC0_LRMS,
                    'ssh'               : gc3libs.defaults.SGE_LRMS,
                    'subprocess'        : gc3libs.defaults.SHELLCMD_LRMS,
        },
    }

    @staticmethod
    def _perform_value_updates(config_items, renames, filename):
        for key, changed in renames.items():
            if key in config_items:
                value = config_items[key]
                if value in changed:
                    gc3libs.log.warning(
                        "Configuration value '%s' was renamed to '%s',"
                        " please change occurrences of '%s=%s' to '%s=%s'"
                        " in configuration file '%s'.",
                        value, changed[value],
                        key, value, key, changed[value],
                        filename)
                    config_items[key] = changed[value]

    _path_key_regexp = re.compile(r'^(\w+_)?(prologue|epilogue)$')

    @staticmethod
    def _perform_filename_conversion(config_items, path_regexp, filename):
        for key, value in config_items.items():
            if path_regexp.match(key):
                basedir = (os.path.dirname(value)
                           if os.path.isfile(value)
                           else os.path.dirname(filename))
                config_items[key] = os.path.join(basedir, value)

    # type transforms for well-known configuration keys
    _convert = {
        # item name             converter
        # ===================   ==================================
        'enabled'             : gc3libs.utils.string_to_boolean,
        'accounting_delay'    : int,
        'architecture'        : _parse_architecture,
        'max_cores'           : int,
        'max_cores_per_job'   : int,
        'max_memory_per_core' : (lambda val: _legacy_parse_memory(val, 'max_memory_per_core')),
        'max_walltime'        : _legacy_parse_duration,
        'override'            : gc3libs.utils.string_to_boolean,
        'port'                : int,
        'vm_os_overhead'      : _legacy_parse_os_overhead,
        'large_file_threshold': (lambda val: _legacy_parse_memory(val, 'large_file_threshold', MB, 'MB')),
        'large_file_chunk_size':(lambda val: _legacy_parse_memory(val, 'large_file_chunk_size', MB, 'MB')),
        # LSF-specific
        'lsf_continuation_line_prefix_length': int,
    }

    @staticmethod
    def _perform_type_conversions(config_items, converters, filename):
        for key, converter in converters.items():
            if key in config_items:
                try:
                    config_items[key] = converter(config_items[key])
                except Exception as err:
                    raise gc3libs.exceptions.ConfigurationError(
                        "Error parsing configuration item '%s': %s: %s"
                        % (key, err.__class__.__name__, err))

    @property
    def auth_factory(self):
        """
        The instance of `gc3libs.authentication.Auth`:class: used to
        manage auth access for the resources.

        This is a *read-only* attribute, created upon first access
        with the values set in `self.auths` and `self.auto_enabled`.
        """
        if self._auth_factory is None:
            try:
                self._auth_factory = gc3libs.authentication.Auth(
                    self.auths, self.auto_enable_auth)
            except Exception as err:
                gc3libs.log.critical(
                    "Failed initializing Auth module: %s: %s",
                    err.__class__.__name__, str(err))
                raise
        return self._auth_factory

    def make_auth(self, name):
        """
        Return factory for auth credentials configured in section
        ``[auth/name]``.
        """
        # use `lambda` for delayed evaluation
        return (lambda **extra_args: self.auth_factory.get(name, **extra_args))

    _removed_resource_types = (
        gc3libs.defaults.ARC0_LRMS,
        gc3libs.defaults.ARC1_LRMS,
    )

    # map resource type name (e.g., 'sge' or 'openstack') to module
    # name + class/function within that module
    TYPE_CONSTRUCTOR_MAP = {
        gc3libs.defaults.EC2_LRMS: ("gc3libs.backends.ec2",  "EC2Lrms"),
        gc3libs.defaults.LSF_LRMS: ("gc3libs.backends.lsf",  "LsfLrms"),
        gc3libs.defaults.PBS_LRMS: ("gc3libs.backends.pbs",  "PbsLrms"),
        gc3libs.defaults.OPENSTACK_LRMS:
                                   ("gc3libs.backends.openstack", "OpenStackLrms"),
        gc3libs.defaults.SGE_LRMS: ("gc3libs.backends.sge",  "SgeLrms"),
        gc3libs.defaults.SHELLCMD_LRMS:
                                   ("gc3libs.backends.shellcmd", "ShellcmdLrms"),
        gc3libs.defaults.SLURM_LRMS:
                                   ("gc3libs.backends.slurm", "SlurmLrms"),
    }

    def _get_resource_constructor(self, resource_type):
        """
        Return the callable to be used to instanciate resource of the
        given type name.
        """
        if resource_type not in self._resource_constructors_cache:
            if resource_type not in self.TYPE_CONSTRUCTOR_MAP:
                raise gc3libs.exceptions.ConfigurationError(
                    "Unknown resource type '%s'" % resource_type)
            else:
                modname, clsname = self.TYPE_CONSTRUCTOR_MAP[resource_type]
                try:
                    mod = __import__(modname,
                                     globals(), locals(),
                                     [clsname], 0)
                    cls = getattr(mod, clsname)
                except (ImportError, AttributeError) as err:
                    raise gc3libs.exceptions.Error(
                        ("Could not instanciate"
                         " resource type '{type}': {errcls}: {errmsg}"
                         .format(
                             type=resource_type,
                             errcls=err.__class__.__name__,
                             errmsg=err)),
                        do_log=True)
                self._resource_constructors_cache[resource_type] = cls
                gc3libs.log.debug(
                    "Using class %r from module %r"
                    " to instanciate resources of type %s",
                    cls, mod, resource_type)
                return cls
        return self._resource_constructors_cache[resource_type]

    def make_resources(self, ignore_errors=True):
        """
        Make backend objects corresponding to the configured resources.

        Return a dictionary, mapping the resource name (string) into
        the corresponding backend object.

        By default, errors in constructing backends (e.g., due to a
        bad configuration) are silently ignored: the offending
        configuration is just dropped.  This can be changed by setting
        the optional argument `ignore_errors` to `False`: in this
        case, an exception is raised whenever we fail to construct a
        backend.
        """
        resources = {}
        for name, resdict in self.resources.items():
            try:
                backend = self._make_resource(resdict)
                if backend is None:  # resource is disabled
                    continue
                assert name == backend.name
            except Exception as err:
                # Print the backtrace only if loglevel is DEBUG or
                # more.
                exc_info = gc3libs.log.level <= gc3libs.logging.DEBUG
                gc3libs.log.warning(
                    "Failed creating backend for resource '%s' of type '%s':"
                    " %s: %s",
                    resdict.get(
                        'name',
                        '(unknown name)'),
                    resdict.get(
                        'type',
                        '(unknown type)'),
                    err.__class__.__name__,
                    str(err),
                    exc_info=exc_info)
                if ignore_errors:
                    continue
                else:
                    raise
            resources[name] = backend
        return resources

    def _make_resource(self, resdict):
        """
        Return a backend initialized from the key/value pairs in `resdict`.
        """
        # name' should have been defined by the caller, so if it's
        # missing it's an internal coherency error
        assert 'name' in resdict, (
            "Invalid resource definition '%s': missing required key 'name'."
            % resdict)

        # default values
        resdict.setdefault('enabled', True)
        if not resdict['enabled']:
            gc3libs.log.info(
                "Dropping computational resource '%s'"
                " because of 'enabled=False' setting"
                " in configuration file.",
                resdict['name'])
            return None

        # minimal sanity check
        for key in self._resource_required_keys:
            if key not in resdict:
                raise gc3libs.exceptions.ConfigurationError(
                    "Missing required parameter '{key}'"
                    " in definition of resource '{name}'."
                    .format(key=key, name=resdict['name']))
        for key in self._resource_keys_value_required:
            if key in resdict and not resdict[key]:
                raise gc3libs.exceptions.ConfigurationError(
                    "Found empty configuration key `{key}`"
                    " in definition of resource '{name}'."
                    .format(key=key, name=resdict['name']))

        if resdict['type'] in self._removed_resource_types:
            resdict['enabled'] = False
            gc3libs.log.warning(
                "Dropping computational resource '{name}':"
                " resource type '{type}' is no longer supported."
                " Please update your configuration file."
                .format(**resdict))
            return None

        if __debug__:
            gc3libs.log.debug(
                "Creating resource '%s' defined by: %s.",
                resdict['name'], ', '.join([
                    ("%s=%r" % (k, v)) for k, v in sorted(resdict.items())
                ]))

        for auth_param in 'auth', 'vm_auth':
            if auth_param in resdict:
                resdict[auth_param] = self.make_auth(resdict[auth_param])

        try:
            # valid strings can be, e.g., `shellcmd+ssh` or `sge`
            resource_type = resdict['type'].split('+')[0]
            cls = self._get_resource_constructor(resource_type)
            # check that required parameters are given, and try to
            # give a sensible error message if not; if we do not
            # do this, users see a traceback like this::
            #
            #   gc3.gc3libs: ERROR: Could not create resource \
            #       'schroedinger-via-ssh': __init__() takes at least 10 \
            #       non-keyword arguments (9 given). Configuration file \
            #       problem?
            #
            # which gives no clue about what to correct!
            args, varargs, keywords, defaults = inspect.getargspec(
                cls.__init__)
            if defaults is not None:
                # `defaults` is a list of default values for the last N args
                defaulted = dict((argname, value)
                                 for argname, value in zip(reversed(args),
                                                           reversed(defaults)))
            else:
                # no default values at all
                defaulted = {}
            for argname in args[1:]:  # skip `self`
                if argname not in resdict and argname not in defaulted:
                    raise gc3libs.exceptions.ConfigurationError(
                        "Missing required configuration parameter '%s'"
                        " for resource '%s'" % (argname, resdict['name']))

            # finally, try to construct backend class...
            return cls(**dict(resdict))

        except Exception as err:
            gc3libs.log.error(
                "Could not create resource '%s': %s. Configuration file"
                " problem?" % (resdict['name'], str(err)))
            raise


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="config",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gsinusrange.py -- Front-end script for running Matlab function
#   function over a large parameter range.
#
#   Copyright (C) 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for running Matlab function
#   function over a large parameter range.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsinusrange.py --help`` for program usage
instructions.

"""

# summary of user-visible changes
from __future__ import absolute_import, print_function
__changelog__ = """
  2016-09-19:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gsinusrange
    gsinusrange.GsinusrangeScript().run()

import os
import sys
import time
import tempfile

import tarfile
import shutil
import pandas

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_REMOTE_OUTPUT_FOLDER = "./results"
DEFAULT_FUNCTION = "sinusrange"
TARFILE="source.tgz"
TEMP_FOLDER="/var/tmp"


## utility funtions

def _getchunk(file_to_chunk, chunk_size=2, chunk_files_dir='/var/tmp'):
    """
    Takes a file_name as input and a defined chunk size
    ( uses 1000 as default )
    returns a list of filenames 1 for each chunk created and a corresponding
    index reference
    e.g. ('/tmp/chunk2800.csv,2800) for the chunk segment that goes
    from 2800 to (2800 + chunk_size)
    """
    
    chunks = []
    
    # creating 'chunk_files_dir'
    if not(os.path.isdir(chunk_files_dir)):
        try:
            os.mkdir(chunk_files_dir)
        except OSError, osx:
            gc3libs.log.error("Failed while creating tmp folder %s. " % chunk_files_dir +
                              "Error %s." % str(osx) +
                              "Using default '/tmp'")
            chunk_files_dir = "/tmp"

    reader = pandas.read_csv(file_to_chunk, header=0, chunksize=chunk_size)
    
    index = 0
    for chunk in reader:
        index += 1     
        filename = "%s/chunk_%s.csv" % (chunk_files_dir,index)
        chunk.to_csv(filename, header=True, index=False)
        chunks.append((filename,index))
            
    return chunks


def _scanandtar(dir_to_scan, temp_folder=TEMP_FOLDER):
    try:
        gc3libs.log.debug("Compressing input folder '%s'" % dir_to_scan)
        cwd = os.getcwd()
        os.chdir(dir_to_scan)

        if not os.path.isdir(temp_folder):
            os.mkdir(temp_folder)
        
        with tarfile.open(os.path.join(temp_folder,TARFILE), "w:gz") as tar:

            tar.add(dir_to_scan, arcname=".")
            os.chdir(cwd)

            gc3libs.log.info("Created tar file '%s'" % TARFILE)
            return tar.name
        
    except Exception, x:
        gc3libs.log.error("Failed creating input archive '%s': %s %s",
                          os.path.join(dir_to_scan,),
                          type(x),x.message)
        raise

    
## custom application class

class GsinusrangeApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gsinusrange'
    
    def __init__(self, input_file, mfunct, **extra_args):


        executables = []
        inputs = dict()
        outputs = dict()

        wrapper = resource_filename(Requirement.parse("gc3pie"),
                                    "gc3libs/etc/gthermostat.sh")
        inputs[wrapper] = "./wrapper.sh"

        inputs[input_file] = os.path.basename(input_file)

        arguments = "./wrapper.sh %s %s %s" % (mfunct,
                                               os.path.basename(input_file),
                                               DEFAULT_REMOTE_OUTPUT_FOLDER)

        if 'source' in extra_args:
            inputs[extra_args['source']] = os.path.basename(extra_args['source'])
            arguments += " -s %s" % os.path.basename(extra_args['source'])
            
        # Set output
        outputs[DEFAULT_REMOTE_OUTPUT_FOLDER] = DEFAULT_REMOTE_OUTPUT_FOLDER

        gc3libs.log.debug("Creating application for executing: %s",
                          arguments)
        
        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gsinusrange.log',
            join=True,
            executables = executables,
            **extra_args)

class GsinusrangeScript(SessionBasedScript):
    """
    Takes 1 .csv input file containing the list of parameters to be passed
    to the `sinusrange` application.
    Each line of the input .csv file correspond to the parameter list to be
    passed to a single `ctx-linkdyn-ordprm-sirs.p4` execution. For each line
    of the input .csv file a GsinusrangeApplication needs to be generated (depends on
    chunk value passed as part of the input options).
    Splits input .csv file into smaller chunks, each of them of size 
    'self.params.chunk_size'.
    Then submits one execution for each of the created chunked files.

    The ``gsinusrange`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.
    
    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsinusrange`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gsinusrange``
    aggregates them into a single larger output file located in 
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GsinusrangeApplication, 
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GsinusrangeApplication,
            )

    def setup_args(self):
        
        self.add_param('input_ranges', type=str,
                       help="folder containing input ranges .txt files")

    def setup_options(self):
        self.add_param("-R", "--src", metavar="[STRING]", 
                       dest="source", default=None,
                       help="Location of the Matlab functions.")

        self.add_param("-f", "--function", metavar="STRING",
                       dest="mfunct", default=DEFAULT_FUNCTION,
                       help="Name of the Matlab function to call."
                       " Default: %(default)s.")

        
    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """
        try:
            assert os.path.isdir(os.path.abspath(self.params.input_ranges)),  \
                "Input ranges folder not found: %s " \
                % os.path.abspath(self.params.input_ranges)

            if self.params.source:
                assert os.path.isdir(os.path.abspath(self.params.source)),  \
                    "Matlab source folder not found: %s " \
                    % os.path.abspath(self.params.source)

            assert os.path.isfile(os.path.join(self.params.source,self.params.mfunct + ".m")), \
                "Matlab funtion file %s/%s.m not found" % (self.params.source,
                                                           self.params.mfunct)
            
        except AssertionError as ex:
            raise ValueError(ex.message)            

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new Application
        """
        tasks = []

        if self.params.source:
            tar_file = _scanandtar(os.path.abspath(self.params.source),
                                  temp_folder=os.path.join(self.session.path,"tmp"))

        for input_range_file in os.listdir(self.params.input_ranges):
            
            jobname = "gsinusrange-%s" % os.path.basename(input_range_file)

            extra_args = extra.copy()

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            if self.params.source:
                extra_args['source'] = tar_file
            tasks.append(GsinusrangeApplication(
                os.path.join(self.params.input_ranges,input_range_file),
                self.params.mfunct,
                **extra_args))

        return tasks



#! /usr/bin/env python
#
"""
Unit tests for the OpenStack backend
"""
# Copyright (C) 2012-2013, 2015  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

# stdlib imports
from collections import namedtuple
import os

# 3rd party imports
from mock import MagicMock
import pytest

# local imports
from gc3libs import Application
import gc3libs.config
import gc3libs.exceptions
from gc3libs.quantity import MiB, GB

# The OpenStack backend might not be installed (it's currently marked as
# optional in `setup.py`), or we may be running Python 2.6 which is no longer
# supported so skip these tests altogether if there is any error
OpenStackLrms = pytest.importorskip('gc3libs.backends.openstack').OpenStackLrms


from gc3libs.testing.helpers import temporary_config, temporary_config_file

class _const(object):
    SSH_AUTH_STANZA = """
[auth/gc3user_ssh]
type = ssh
username = gc3-user

"""

    RESOURCE_STANZA = """
[resource/hobbes]
os_region=RegionOne
public_key=~/.ssh/id_dsa.pub
name=hobbes
enabled = yes
type=openstack+shellcmd
auth=hobbes
max_cores_per_job = 8
max_memory_per_core = 2
max_walltime = 8
max_cores = 32
architecture = x86_64
keypair_name=keypair
vm_auth=gc3user_ssh
image_id=a9e98055-5aa3-4636-8fc5-f3b2b4ea66bb
security_group_name=gc3pie_ssh
security_group_rules=tcp:22:22:0.0.0.0/0, icmp:-1:-1:0.0.0.0/0

"""

    CLOUD_AUTH_STANZA_FULL = """
[auth/hobbes]
type=openstack
os_username=USERNAME
os_password=PASSWORD
os_project_name=TENANT

"""

    CLOUD_AUTH_STANZA_USE_ENV = """
[auth/hobbes]
type=openstack
# other fields should be set via environment variables

"""

@pytest.mark.xfail
def test_openstack_variables_are_optional():
    with temporary_config_file(
        _const.SSH_AUTH_STANZA
        + _const.CLOUD_AUTH_STANZA_USE_ENV
        + _const.RESOURCE_STANZA
    ) as cfgfile:
        # set up fake environment
        env = ['username', 'password', 'tenant_name', 'auth_url']
        for name in env:
            os.environ['OS_' + name.upper()] = name

        # check that resource has been correctly created
        cfg = gc3libs.config.Configuration(cfgfile.name)
        resources = cfg.make_resources()
        assert 'hobbes' in resources

        cloud = resources['hobbes']
        assert isinstance(cloud, OpenStackLrms)

        # check that values have been inherited from the environment
        for name in env:
            attr = ('os_' + name)
            assert hasattr(cloud, attr)
            assert getattr(cloud, attr) == name


def _setup_flavor_selection_tests(extra_conf=''):
    # FIXME: the test does not run if this is not in the environment!
    os.environ['OS_AUTH_URL'] = 'http://localhost:5000/v2/'
    env = ['username', 'password', 'tenant_name', 'auth_url']
    for name in env:
        os.environ['OS_' + name.upper()] = name

    with temporary_config_file(
        _const.SSH_AUTH_STANZA
        + _const.CLOUD_AUTH_STANZA_FULL
        + _const.RESOURCE_STANZA + extra_conf
    ) as cfgfile:
        cfg = gc3libs.config.Configuration(cfgfile.name)
        resources = cfg.make_resources()
        cloud = resources['hobbes']

    # mock novaclient's `Flavor` type using a namedtuple
    _Flavor = namedtuple('_Flavor', 'name vcpus ram disk'.split())
    flavors = (
        _Flavor(name='1cpu-4ram-hpc', vcpus=1, ram=4000, disk=100),
        _Flavor(name='4cpu-16ram-hpc', vcpus=4, ram=16000, disk=100),
        _Flavor(name='8cpu-8ram-server', vcpus=8, ram=7680, disk=100),
    )

    cloud.client = MagicMock()
    cloud.client.flavors.list.return_value = flavors
    # `_connect` fills the flavor list
    cloud._connect()

    return cloud, flavors


@pytest.mark.xfail
def test_flavor_selection_at_init():
    """
    Test that the "largest" flavor is correctly identified.
    """
    cloud, flavors = _setup_flavor_selection_tests()

    # check selection of the largest flavor in lexicographic order
    # (cpus, ram, disk)
    assert cloud['max_cores'] == flavors[2].vcpus
    assert cloud['max_memory_per_core'] == flavors[2].ram * MiB


@pytest.mark.xfail
def test_flavor_selection_default():
    """
    Test flavor-selection logic when no "instance type" is specified.
    """
    cloud, flavors = _setup_flavor_selection_tests()

    # no instance type specified, should get smallest one
    app1 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    flv = cloud.get_instance_type_for_job(app1)
    assert flv == flavors[0]

    # require 10 GB, get 4cpu-16ram-hpc flavor
    app2 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
        requested_memory=10*GB,
    )
    flv = cloud.get_instance_type_for_job(app2)
    assert flv == flavors[1]

    # require 6 cores, get 8cpu-8ram-server flavor
    app3 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
        requested_cores=6,
    )
    flv = cloud.get_instance_type_for_job(app3)
    assert flv == flavors[2]


@pytest.mark.xfail
def test_flavor_selection_generic_instance_type():
    """
    Test flavor selection when a default flavor is prescribed by config.
    """
    # NOTE: `EXTRA_CONF` has to start with a blank line and with the first
    # non-whitespace character in column 0, otherwise `ConfigParser` will
    # consider it a continuation of the last key/value assignment in
    # `RESOURCE_STANZA` ...
    EXTRA_CONF = """

instance_type = 4cpu-16ram-hpc
small_instance_type = 1cpu-4ram-hpc
"""
    cloud, flavors = _setup_flavor_selection_tests(EXTRA_CONF)

    # no instance type specified, should get configured one
    app1 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    flv = cloud.get_instance_type_for_job(app1)
    assert flv == flavors[1]

    # application-specific instance type specified, should override default
    app2 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    app2.application_name = 'small'
    flv = cloud.get_instance_type_for_job(app2)
    assert flv == flavors[0]

    # requirements exclude default instance type
    app3 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
        requested_cores=6,
    )
    app3.application_name = 'large'
    flv = cloud.get_instance_type_for_job(app3)
    assert flv == flavors[2]


@pytest.mark.xfail
def test_flavor_selection_application_specific():
    """
    Test flavor selection when an application-specific flavor is prescribed by config.
    """
    # NOTE: `EXTRA_CONF` has to start with a blank line and with the first
    # non-whitespace character in column 0, otherwise `ConfigParser` will
    # consider it a continuation of the last key/value assignment in
    # `RESOURCE_STANZA` ...
    EXTRA_CONF = """

large_instance_type = 4cpu-16ram-hpc
"""
    cloud, flavors = _setup_flavor_selection_tests(EXTRA_CONF)

    # no instance type specified, should get smallest one
    app1 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    flv = cloud.get_instance_type_for_job(app1)
    assert flv == flavors[0]

    # application-specific instance type specified
    app2 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    app2.application_name = 'large'
    flv = cloud.get_instance_type_for_job(app2)
    assert flv == flavors[1]

    # application-specific instance type specified,
    # but requirements exclude it
    app3 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
        requested_cores=6,
    )
    app3.application_name = 'large'
    flv = cloud.get_instance_type_for_job(app3)
    assert flv == flavors[2]


@pytest.mark.xfail
def test_flavor_selection_application_specific_unavailable():
    """
    Test that application-specific "instance type" is ignored if unavailable.
    """
    # NOTE: `EXTRA_CONF` has to start with a blank line and with the first
    # non-whitespace character in column 0, otherwise `ConfigParser` will
    # consider it a continuation of the last key/value assignment in
    # `RESOURCE_STANZA` ...
    EXTRA_CONF = """

large_instance_type = 5cpu-11ram-server
"""
    cloud, flavors = _setup_flavor_selection_tests(EXTRA_CONF)

    # no instance type specified, should get smallest one
    app1 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    flv = cloud.get_instance_type_for_job(app1)
    assert flv == flavors[0]

    # application-specific instance type specified but unavailable:
    # should yield same result as before
    app2 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
    )
    app2.application_name = 'large'
    flv = cloud.get_instance_type_for_job(app2)
    assert flv == flavors[0]

    # application-specific instance type specified,
    # but requirements exclude it
    app3 = Application(
        ['/bin/true'],
        inputs=[],
        outputs=[],
        output_dir='/tmp',
        requested_cores=6,
    )
    app3.application_name = 'large'
    flv = cloud.get_instance_type_for_job(app3)
    assert flv == flavors[2]


# main: run tests

if "__main__" == __name__:
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gminarevix.py -- Front-end script for running Matlab MinAREVIX_Cloud
#   function over a set of structure data and structure models.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gminarevix.py --help`` for program usage
instructions.

Input parameters consists of:

...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2014-03-14:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gminarevix
    gminarevix.GminarevixScript().run()

import os
import sys
import time
import tempfile

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

# Resource requirements depending on Model type (only # cores)
MODELS_SPECS = {'Wishart': {'requested_cores': 8, 'requested_memory': Memory(12*GB)},
                'SVJJ': {'requested_cores': 16, 'requested_memory': Memory(32*GB)},
                'SV2F': {'requested_cores': 8, 'requested_memory': Memory(12*GB)}}

## custom application class
class GminarevixApplication(Application):
    """
    Custom class to wrap the execution of the Matlab script.
    Binary is called 'minarevix_cloud'; it can be replaced by what is passed as
    'run_binary' option.
    Binary takes a single 'model_name' and a data folder containing 1 structure_data
    in .mat format.
    Output will be written in 'data' folder (this is hardcoded in the Matlab
    differential evolutionary alghorithm.
    """
    application_name = 'matlab-mcr'

    def __init__(self, model_name, structure_data, **extra_args):

        self.output_dir = extra_args['output_dir']

        inputs = dict()
        outputs = dict()

        data_name = os.path.basename(structure_data)
        inputs[structure_data] = "./data/%s" % data_name
        outputs["./data/"] = "results/"

        if extra_args.has_key('run_binary'):
            inputs[os.path.abspath(extra_args['run_binary'])] = './minarevix_cloud'
            arguments = "./minarevix_cloud "
        else:
            arguments = "minarevix_cloud "

        arguments += "%s ./data/%s" % (model_name, data_name)

        if extra_args.has_key('requested_cores'):
            extra_args['requested_cores'] = max(MODELS_SPECS[model_name]['requested_cores'],
                                                extra_args['requested_cores'])
        else:
            extra_args['requested_cores'] = MODELS_SPECS[model_name]['requested_cores']

        if extra_args.has_key('requested_memory'):
            extra_args['requested_memory'] = max(MODELS_SPECS[model_name]['requested_memory'],
                                                 extra_args['requested_memory'])
        else:
            extra_args['requested_memory'] = MODELS_SPECS[model_name]['requested_memory']

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gminarevix.log',
            join=True,
            **extra_args)

class GminarevixScript(SessionBasedScript):
    """
    Fro each structure_data file (with '.mat' extension) found in the 'input folder',
    GminarevixScript generates as many Tasks as 'models' defined.

    The ``gminarevix`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gminarevix`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GminarevixApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GminarevixApplication,
            )

    def setup_args(self):


        self.add_param('models', type=str,
                       help="Comma separated list of model names. "
                       "Allowed models: %s" % str(MODELS_SPECS.keys()))

        self.add_param('structure_data', type=str,
                       help="Path to the folder containing structure data files.")

    def setup_options(self):
        self.add_param("-b", "--binary", metavar="[STRING]",
                       dest="run_binary", default=None,
                       help="Location of the Matlab compiled binary "
                       "version of the MinArevix.")

    def parse_args(self):
        """
        Check validity of input parameters and selected benchmark.
        """

        if self.params.run_binary:
            if not os.path.isfile(self.params.run_binary):
                raise gc3libs.exceptions.InvalidUsage("MinArevix binary "
                                                      " file %s not found"
                                                      % self.params.run_binary)


        # check args:
        if not os.path.isdir(self.params.structure_data):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to structure data: '%s'. Folder not found"
                % self.params.structure_data)

        for model in self.params.models.split(','):
            if not model in MODELS_SPECS.keys():
                raise gc3libs.exceptions.InvalidUsage(
                    "Invalid model name: %s. Authorized model names: %s"
                    % (model, str(MODELS_SPECS.keys())))
            else:
                gc3libs.log.info("valid model: %s", model)

    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GminarevixApplication.

        First loop the input files, then loop the selected benchmarks
        """
        tasks = []

        for input_file_name in os.listdir(self.params.structure_data):

            # Rule-out files without '.dat' extension
            if not input_file_name.endswith(".mat"):
                continue

            input_file = os.path.abspath(os.path.join(self.params.structure_data, input_file_name))

            for model in self.params.models.split(','):
                # extract numerical value from filename given the structure:
                # VIX_Call_Options_15102008.mat
                (a,b,c,data_index) = input_file_name.split('_')

                # XXX: need to find a more compact name
                # jobname = "gminarevix-%s-%s" % (model,data_index)
                jobname = "gminarevix-%s-%s" % (model,(input_file_name))

                extra_args = extra.copy()

                extra_args['jobname'] = jobname
                # extra_args['model_name'] = model

                if self.params.run_binary:
                    extra_args['run_binary'] = self.params.run_binary

                extra_args['output_dir'] = self.params.output
                extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                            os.path.join(model,
                                                                                         input_file_name))
                extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                            os.path.join(model,
                                                                                         input_file_name))
                extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                            os.path.join(model,
                                                                                         input_file_name))
                extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                            os.path.join(model,
                                                                                         input_file_name))

                tasks.append(GminarevixApplication(
                    model,
                    input_file,
                    **extra_args))

        return tasks



#! /usr/bin/env python
#
from __future__ import absolute_import, print_function, unicode_literals
"""
"""
# Copyright (C) 2011-2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Check if all the backends are implementing all the needed methods.
"""

__docformat__ = 'reStructuredText'

from gc3libs.backends import LRMS


def check_class(cls):
    for name in [  # list of abstract methods in class `LRMS`
            'cancel_job',
            'close',
            'free',
            'get_resource_status',
            'get_results',
            'peek',
            'submit_job',
            'update_job_state',
            'validate_data',
    ]:
        if getattr(cls, name) == getattr(LRMS, name):
            raise NotImplementedError(
                "Abstract method `%s` not implemented in class `%s`"
                % (name, cls.__name__))


def test_shellcmd_backends():
    from gc3libs.backends.shellcmd import ShellcmdLrms
    check_class(ShellcmdLrms)


def test_lsf_backends():
    from gc3libs.backends.lsf import LsfLrms
    check_class(LsfLrms)


def test_pbs_backends():
    from gc3libs.backends.pbs import PbsLrms
    check_class(PbsLrms)


def test_sge_backends():
    from gc3libs.backends.sge import SgeLrms
    check_class(SgeLrms)


if "__main__" == __name__:
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gpyrad.py -- Front-end script for submitting multiple `PyRAD` jobs.
#
#   Copyright (C) 2013, 2014  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `PyRAD` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gpyrad --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """

  2014-04-16:
  * Retrieve only specific result folder as output.
  * Initial experimental support for S3 repository

  2014-02-24:
  * Initial release, forked off the ``ggeosphere`` sources.
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gpyrad
    gpyrad.GpyradScript().run()

from pkg_resources import Requirement, resource_filename
import os
import posix

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds, GiB
from gc3libs.workflow import RetryableTask
import gc3libs.url

# Default values
DEFAULT_WCLUST="0.9"

## custom application class

class GpyradApplication(Application):
    """
    Fetches execution wrapper, input file and checks
    whether optional arguments have been passed.
    Namely ''wclust'' that sets the pyRAD clustering
    treshold and ''paramsfile'' needed to run PyRAD.
    """

    application_name = 'pyrad'

    def __init__(self, input_file, **extra_args):
        """
        The wrapper script is being used for start the simulation.
        """

        inputs = []
        outputs = []

        gpyrad_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gpyrad_wrapper.sh")


        inputs.append((gpyrad_wrapper_sh,os.path.basename(gpyrad_wrapper_sh)))

        cmd = "./gpyrad_wrapper.sh "

        if 's3cfg' in extra_args:
            inputs.append((extra_args['s3cfg'],
                           "etc/s3cfg"))

        if 'wclust' in extra_args:
            cmd += " -w %s " % extra_args['wclust']
            output_folder_name = 'clust%s' % extra_args['wclust']
        else:
            # This is a convention of PyRAD
            output_folder_name = 'clust%.1f' % DEFAULT_WCLUST
        outputs.append(output_folder_name)

        if 'debug' in extra_args:
            cmd += "-d "
            outputs.append('strace.log')

        if 'paramsfile' in extra_args:
            cmd += " -p ./params.tmpl "
            #XXX: params file contains important paths needed
            # by pyRAD. If we deploy an alternative params.txt file supplied
            # by the end-user, we might incurr in the risk that we can no longer use
            # the assumptions we made with the original params.txt file.
            inputs.append((extra_args['paramsfile'],'./params.tmpl'))

        remote_input_file = os.path.join('./input',os.path.basename(input_file))

        cmd += " %s " % remote_input_file
        inputs.append((input_file,remote_input_file))

        # Add memory requirement
        extra_args.setdefault('requested_memory', 1.5*GiB)

        Application.__init__(
            self,
            # arguments should mimic the command line interfaca of the command to be
            # executed on the remote end
            arguments = cmd,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gpyrad.log',
            join=True,
            **extra_args)

class GpyradTask(RetryableTask, gc3libs.utils.Struct):
    """
    Run ``pyrad`` on a given simulation directory until completion.
    """
    def __init__(self, input_file, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GpyradApplication(input_file, **extra_args),
            # keyword arguments
            **extra_args)

## main script class

class GpyradScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for simulation
directories and submit a job for each one found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

The ``gpyrad`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gpyrad`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GpyradTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GpyradTask,
            )


    def setup_options(self):
        self.add_param("-T", "--treshold", metavar="INT",
                       dest="wclust", default=None,
                       help="Wclust: clustering threshold as a decimal."+
                       "Note: if '-p' is used, this option will be ignored.")

        self.add_param("-P", "--params", metavar="PATH",
                       dest="paramsfile", default=None,
                       help="Location of params.txt fiule required by pyrad. "+
                       "WARNING: The default params.txt is used to make "+
                       "assumption on the location of pyRAD and the "+
                       "input folder where the *.fastq files are located. "+
                       "Passing an alternative params.txt file might break "+
                       "such assumptions resulting in a falied execution. " +
                       "Please check the default params.txt file located in "+
                       "'test.pyRAD' folder where gpyrad.py is located.")


        self.add_param("-Y", "--s3cfg", metavar="PATH",
                       dest="s3cfg", default=None,
                       help="Location of the s3cfg configuration file. "+
                       "Default assumes 's3cfg' configuration file being "+
                       "available on the remote execution node.")

    def setup_args(self):

        self.add_param('input_container', type=str,
                       help="Path to local folder containing fastq files.")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        self.input_folder_url = gc3libs.url.Url(self.params.input_container)

        if self.params.paramsfile and not os.path.isfile(self.params.paramsfile):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid params file '%s': File not found"
                % self.params.paramsfile)

        if self.params.s3cfg and not os.path.isfile(self.params.s3cfg):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid S3cmd config file '%s': File not found"
                % self.params.s3cfg)

    def new_tasks(self, extra):

        tasks = []

        #XXX: I would have like to use the same forumla as in
        # cmdline:405 but I found it very confusing
        # easier to use a linear threshold
        # 0: error 1: warning 2: info 3: debug
        if self.params.verbose > 3:
            extra['debug'] = True

        for input_file in self._list_folder_by_url(self.input_folder_url):

            jobname = "%s" % os.path.basename(input_file).split(".")[0]

            extra_args = extra.copy()
            extra_args['jobname'] = jobname

            # FIXME: ignore SessionBasedScript feature of customizing
            # output folder
            extra_args['output_dir'] = self.params.output

            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

            if self.params.paramsfile:
                extra_args['paramsfile'] = self.params.paramsfile
            elif self.params.wclust:
                extra_args['wclust'] = self.params.wclust

            if self.params.s3cfg:
                extra_args['s3cfg'] = self.params.s3cfg

            self.log.info("Creating Task for input file: %s" % input_file)

            tasks.append(GpyradTask(
                input_file,            # path to valid input file
                **extra_args
                ))

        return tasks


    def _list_folder_by_url(self, url):
        """
        """

        if url.scheme == "s3":
            return self._list_s3_container(url)
        elif url.scheme ==  "file":
            return self._list_local_folder(url)
        else:
            gc3libs.log.error("Unsupported Input folder URL %s. "+
                              "Only supported protocols are: [file,s3] " % url.scheme)
            return None

    def _list_local_folder(self, input_folder):
        """
        return a list of all .fastq files in the input folder
        """

        return [ os.path.join(input_folder.path,infile) for infile in os.listdir(input_folder.path) if infile.endswith('.fastq') ]

    def _list_S3_container(self, s3_url):
        """
        Use s3cmd command line interface to interact with
        a remote S3-compatible ObjectStore.
        Assumption:
        . s3cmd configuration file available
        and correctly pointing to the right ObjectStore.
        . s3cmd available in PATH environmental variable.
        . Valid for only 1 S3_URL path
        """

        import subprocess

        # read content of remote S3CMD_URL
        try:
            # 's3cmd ls' should return a list of model archives
            # for each of them bundle archive_name and working_dir
            _process = subprocess.Popen("s3cmd ls %s" % s3_url,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        close_fds=True, shell=True)

            (out, err) = _process.communicate()
            exitcode = _process.returncode

            if (exitcode != 0):
                raise Exception("Error: %s, %s", (out,err))

            # Parse 'out_list' result and extract all .tgz or .zip archive names

            for s3_obj in out.strip().split("\n"):
                if s3_obj.startswith("DIR"):
                    # it's a S3 directory; ignore
                    continue
                # object string format: '2014-01-09 16:41 3627374 s3://a4mesh/model_1.zip'
                s3_url = s3_obj.split()[3]
                if(s3_url.endswith(".fastq")):
                   yield s3_url

        except Exception, ex:
            gc3libs.log.error("Failed while reading remote S3 container. "+
                              "%s", ex.message)



#!/usr/bin/env python

"""
export SUBJECTS_DIR=/our/out/data/path

************
SUBJECT: s01
************
 ************
 DATA INPUT s01
 ************
recon-all -i in_data_path/s01/TP1/T1w_a.nii.gz -i in_data_path/s01/TP1/T1w_b.nii.gz -subjid s01.cross.TP1
recon-all -i in_data_path/s01/TP2/T1w_a.nii.gz -i in_data_path/s01/TP2/T1w_b.nii.gz -subjid s01.cross.TP2
recon-all -i in_data_path/s01/TP3/T1w_a.nii.gz -i in_data_path/s01/TP3/T1w_b.nii.gz -subjid s01.cross.TP3

 ************
 CROSS SECTIONAL PROCESSING s01
 ************
recon-all -s s01.cross.TP1 -all
recon-all -s s01.cross.TP2 -all
recon-all -s s01.cross.TP3 -all

 ************
 BASE PROCESSING s01
 ************
recon-all -base s01.base -tp s01.cross.TP1 -tp s01.cross.TP2 -tp s01.cross.TP3 -all

 ************
 LONG PROCESSING s01
 ************
recon-all -long s01.cross.TP1 s01.base -all -qcache
recon-all -long s01.cross.TP2 s01.base -all -qcache
recon-all -long s01.cross.TP3 s01.base -all -qcache

XXX: Allow differnt type of NII extensions. e.g. nii.tgz
XXX: How to formally verify correctness of output from step 'cross'.
     For the itme being do nothing on this.
XXX: Remove simlinks from output folder once completed.
"""

from __future__ import absolute_import, print_function, unicode_literals
import sys
import os
import subprocess

NII_EXTENSION = "nii"
FSAVERAGE = "fsaverage"
FS_SUBJECT_FSAVERAGE = os.path.join(os.environ.get("FREESURFER_HOME", ''),"subjects",FSAVERAGE)

def Usage():
    print ("Usage: gnift_wrapper.py <subject name> <NIFIT input folder> <subjects dir>")

def RunFreesurfer(subject, input, output):
    """
    Walk through `input` and search for every subfolder with at least 1 `nii` file.
    Record reference to `nii` files per subfolder.
    Run Data Input, Cross Sectional, Base and Long for all of them.
    Use only 1 core, run the step sequentially.
    """

    # Create output folder and add simlink to FSAVERAGE
    os.mkdir(output)
    os.symlink(FS_SUBJECT_FSAVERAGE,os.path.join(output,FSAVERAGE))

    input_nii = dict()
    cross_files = []
    os.environ["SUBJECTS_DIR"] = output

    for r,d,f in os.walk(input):
        for filename in f:
            if filename.endswith(NII_EXTENSION):
                timepoint = os.path.basename(r)
                if not timepoint in list(input_nii.keys()):
                    # Initialise
                    input_nii[timepoint] = list()
                input_nii[timepoint].append(os.path.join(r,filename))


    # DATA INPUT and  CROSS SECTIONAL PROCESSING
    print("Start DATA INPUT on %d Timepoints" % len(list(input_nii.keys())))
    for timepoint in sorted(input_nii.keys()):
        inputs = '-i '+' -i '.join(x for x in input_nii[timepoint])
        data_input_outputfolder =  "%s.cross.%s" % (subject, timepoint)

        "Example: recon-all -i in_data_path/s01/TP1/T1w_a.nii.gz -i in_data_path/s01/TP1/T1w_b.nii.gz -subjid s01.cross.TP1"
        command="recon-all -deface %s -subjid %s" % (inputs, data_input_outputfolder)
        runme(command)

        # Verify output
        try:
            assert os.path.isdir(os.path.join(output,data_input_outputfolder)), \
            "Output folder '%s' not found" % os.path.join(output,data_input_outputfolder)
        except AssertionError as ex:
            print(ex.message)
            return 1

        cross_files.append(data_input_outputfolder)

        print("Start CROSS SECTIONAL PROCESSING")
        "Example: recon-all -s s01.cross.TP1 -all"
        command="recon-all -deface -s %s -all" % data_input_outputfolder
        runme(command)


    # BASE PROCESSING
    print("Start BASE PROCESSING with %d timepoints" % len(cross_files))
    inputs = '-tp '+' -tp '.join(x for x in sorted(cross_files))
    basefile = "%s.base" % subject
    command = "recon-all -deface -base %s %s -all" % (basefile,inputs)
    runme(command)
    # XXX: what to verify here ?


    # LONG PROCESSING s01
    print("Start LONG PROCESSING with %d timepoints" % len(cross_files))
    for cross in sorted(cross_files):
        # Example: recon-all -long s01.cross.TP1 s01.base -all -qcache
        command = "recon-all -deface -long %s %s -all -qcache" % (cross,basefile)
        runme(command)
        # XXX: what to verify here ?

def runme(command):
    """
    Comodity function to run commands using `subprocess` module
    Input: command to run
    Output: none
    Raise Exception in case command fails
    """
    proc = subprocess.Popen(
        [command],
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE)

    print("Running command %s" % command)
    (stdout, stderr) = proc.communicate()

    if proc.returncode != 0:
        print("Execution failed with exit code: %d" % proc.returncode)
        print("Output message: %s" % stdout)
        print("Error message: %s" % stderr)
        raise Exeption(stderr)

if __name__ == '__main__':
    if (len(sys.argv) != 4):
        sys.exit(Usage())
    sys.exit(RunFreesurfer(sys.argv[1], sys.argv[2], sys.argv[3]))



#! /usr/bin/env python

"""
Manage startup and teardown of cloud-based VMs to run applications.

This only works on clouds implementing the OpenStack Compute API.
"""

# Copyright (C) 2012-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


import sys
if sys.version_info < (2, 7):
    raise ImportError("GC3Pie's OpenStack backend requires Python 2.7+")

import os
import paramiko
from string import ascii_letters, digits

# OpenStack APIs
try:
    import os_client_config
except ImportError as err:
    from gc3libs.exceptions import ConfigurationError
    raise ConfigurationError(
        "The OpenStack backend is used but the `os_client_config` module"
        " cannot be loaded: {err}. Please, either install it with"
        "`pip install os-client-config` and verify that it works"
        " by running `python -c 'import os_client_config'`,"
        " or update your configuration file and"
        " disable any OpenStack-based resources."
        .format(err=err))

try:
    from novaclient.exceptions import NotFound
except ImportError as err:
    from gc3libs.exceptions import ConfigurationError
    raise ConfigurationError(
        "The OpenStack backend is used but the `novaclient` module"
        " cannot be loaded: {err}. Please, either install it with"
        "`pip install python-novaclient` and verify that it works"
        " by running `python -c 'import novaclient'`,"
        " or update your configuration file and"
        " disable any OpenStack-based resources."
        .format(err=err))


# GC3Pie imports
import gc3libs
import gc3libs.defaults
from gc3libs.exceptions import \
    ConfigurationError, \
    LRMSSubmitError, \
    MaximumCapacityReached, \
    ResourceNotReady, \
    TransportError, \
    UnrecoverableAuthError, \
    UnrecoverableDataStagingError, \
    UnrecoverableError
import gc3libs.url
from gc3libs import Run
from gc3libs.utils import same_docstring_as
from gc3libs.backends import LRMS
from gc3libs.backends.shellcmd import ShellcmdLrms
from gc3libs.backends.vmpool import VMPool, InstanceNotFound
from gc3libs.utils import cache_for, lookup
from gc3libs.quantity import MiB


available_subresource_types = [gc3libs.defaults.SHELLCMD_LRMS]

ERROR_STATES = ['ERROR', 'UNNKNOWN']
PENDING_STATES = ['BUILD', 'REBUILD', 'REBOOT', 'HARD_REBOOT',
                  'RESIZE', 'REVERT_RESIZE']


class OpenStackVMPool(VMPool):
    """
    Implementation of `VMPool` for OpenStack cloud
    """

    def _get_instance(self, vm_id):
        try:
            vm = self.conn.servers.get(vm_id)
        except NotFound:
            raise InstanceNotFound(
                "No instance with id %s has been found." % vm_id)
        return vm


class OpenStackLrms(LRMS):

    """
    OpenStack resource.
    """
    RESOURCE_DIR = '$HOME/.gc3/openstack.d'

    def __init__(self, name,
                 # these parameters are inherited from the `LRMS` class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 # these are specific of the OpenStackLrms class
                 keypair_name, public_key, vm_auth,
                 os_region=None, image_id=None, os_auth_url=None,
                 instance_type=None, auth=None,
                 vm_pool_max_size=None, user_data=None,
                 vm_os_overhead=gc3libs.defaults.VM_OS_OVERHEAD,
                 # extra args are used to instanciate "sub-resources"
                 **extra_args):

        # Note: this creates attributes from key/value pairs given in the
        # `extra_args` parameters. In particular, the `self.type` attribute
        # (referenced below) is set in this chained constructor...
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        self.free_slots = int(max_cores)
        self.user_run = 0
        self.user_queued = 0
        self.queued = 0
        self._flavors = []
        self.vm_pool_max_size = vm_pool_max_size
        if vm_pool_max_size is not None:
            try:
                self.vm_pool_max_size = int(self.vm_pool_max_size)
            except ValueError:
                raise ConfigurationError(
                    "Value for `vm_pool_max_size` must be an integer,"
                    " was %s instead." % vm_pool_max_size)

        # pylint: disable=no-member
        self.subresource_type = self.type.split('+', 1)[1]
        if self.subresource_type not in available_subresource_types:
            raise UnrecoverableError("Invalid resource type: %s" % self.type)

        # Mapping of job.execution.instance_id => LRMS
        self.subresources = {}

        auth = self._auth_fn()
        if os_auth_url is None:
            os_auth_url = os.getenv('OS_AUTH_URL')
        if os_auth_url is None:
            raise gc3libs.exceptions.InvalidArgument(
                "Cannot connect to the OpenStack API:"
                " No 'OS_AUTH_URL' environment variable defined,"
                " and no 'os_auth_url' argument passed"
                " to the OpenStack backend.")
        self.os_auth_url = os_auth_url
        self.os_username = auth.os_username
        self.os_password = auth.os_password
        self.os_tenant_name = auth.os_project_name
        self.os_region_name = os_region
        # Keypair names can only contain alphanumeric chars!
        if not set(keypair_name).issubset(set(ascii_letters + digits + '_-.')):
            raise ConfigurationError(
                "Keypair name `{0}` is invalid:"
                " keypair names can only consist of"
                " alphanumeric chars, plus `_`, `-`, and `.`"
                .format(keypair_name))
        self.keypair_name = keypair_name
        self.public_key = os.path.expanduser(
            os.path.expandvars(public_key.strip()))
        self.image_id = image_id
        self.instance_type = instance_type
        self.user_data = user_data
        self.vm_os_overhead = gc3libs.quantity.Memory(vm_os_overhead)
        self._parse_security_group()
        self._conn = None

        # `*_instance_type` config items should be consumed here,
        # not in any sub-resource
        for key, value in list(extra_args.items()):
            if key.endswith('_instance_type'):
                self[key] = value
                extra_args.pop(key)

        # `self.subresource_args` is used to create subresources
        self.subresource_args = extra_args
        self.subresource_args['type'] = self.subresource_type
        self.subresource_args['architecture'] = self['architecture']
        self.subresource_args['max_cores'] = self['max_cores']
        self.subresource_args['max_cores_per_job'] = self['max_cores_per_job']
        self.subresource_args['max_memory_per_core'] = \
            self['max_memory_per_core']
        self.subresource_args['max_walltime'] = self['max_walltime']
        # SSH-specific configuration
        self.subresource_args['transport'] = 'ssh'
        self.subresource_args['auth'] = vm_auth
        self.subresource_args['ssh_timeout'] = 7  # FIXME: hard-coded!
        self.subresource_args['ignore_ssh_host_keys'] = True
        self.subresource_args['keyfile'] = self.public_key
        if self.subresource_args['keyfile'].endswith('.pub'):
            keyfile = self.subresource_args['keyfile'][:-len('.pub')]
            self.subresource_args['keyfile'] = keyfile
        # ShellcmdLrms by default trusts the configuration, instead of
        # checking the real amount of memory and number of cpus, but
        # we need the real values instead.
        if self.subresource_type == gc3libs.defaults.SHELLCMD_LRMS:
            self.subresource_args['override'] = 'True'

        if image_id is None:
            raise ConfigurationError(
                "No `image_id` specified in the configuration file.")

        # "Connect" to the cloud (connection is actually performed
        # only when needed by the `Client` class.
        self.compute_client, self.image_client, self.network_client = self._new_clients()

        # Set up the VMPool persistent class. This has been delayed
        # until here because otherwise self._conn is None
        pooldir = os.path.join(
            os.path.expandvars(OpenStackLrms.RESOURCE_DIR),
            'vmpool', self.name)
        self._vmpool = OpenStackVMPool(pooldir, self.compute_client)

    def _new_clients(self):
        self.__connected = False
        for kind in ['compute', 'image', 'network']:
            yield os_client_config.make_client(
                kind,
                auth_url=self.os_auth_url,
                username=self.os_username,
                password=self.os_password,
                project_name=self.os_tenant_name,
                region_name=self.os_region_name,
            )

    def _connect(self):
        if not self.__connected:
            self.compute_client.authenticate()
            # Fill the flavors and update the resource.
            self._flavors = self.compute_client.flavors.list()
            flavor = max(self._flavors, key=self._flavor_ordering_key)
            # FIXME: if the configured values are lower, should they be honored instead?
            self['max_cores'] = self['max_cores_per_job'] = flavor.vcpus
            self['max_memory_per_core'] = flavor.ram * MiB
            gc3libs.log.info(
                "Largest VM flavor available on the cloud: %s (%d vCPUs, %s RAM)",
                flavor.name, self['max_cores'], self['max_memory_per_core'])
        self.__connected = True

    def _create_instance(self, image_id, name='gc3pie-instance',
                         instance_type=None, user_data=None):
        """
        Create an instance using the image `image_id` and instance
        type `instance_type`. If no `instance_type` is defined, use
        the default.

        This method will also setup the keypair and the security
        groups, if needed.
        """

        args = {}
        if user_data:
            args['userdata'] = user_data

        # Check if the desired keypair is present
        try:
            keypair = self._get_keypair(self.keypair_name)
        except NotFound:
            gc3libs.log.info(
                "Keypair `%s` not found: creating it using public key `%s`"
                % (self.keypair_name, self.public_key))
            # Create keypair if it does not exist and give an error if it
            # exists but have different fingerprint
            self._import_keypair()
        else:
            self._have_keypair(keypair)
        instance_type = instance_type or self.instance_type
        # Setup security groups
        if 'security_group_name' in self:
            self._setup_security_groups()
            args['security_groups'] = [self.security_group_name]

        # FIXME: we should add check/creation of proper security
        # groups

        nics = None
        if hasattr(self,'network_ids') and self.network_ids:
            nics = [{'net-id': netid.strip(), 'v4-fixed-ip': ''}
                  for netid in self.network_ids.split(',')]
            gc3libs.log.debug("Specifying networks for vm %s: %s",
                      name, ', '.join([nic['net-id'] for nic in nics]))
        args['nics'] = nics

        gc3libs.log.debug("Create new VM using image id `%s`", image_id)
        try:
            vm = self.compute_client.servers.create(name, image_id, instance_type,
                                            key_name=self.keypair_name, **args)
        except Exception as err:
            # scrape actual error kind and message out of the
            # exception; we do this mostly for sensible logging, but
            # could be an actual improvement to Boto to provide
            # different exception classes based on the <Code>
            # element...
            # XXX: is there a more robust way of doing this?
            # fall back to normal reporting...
            raise UnrecoverableError("Error starting instance: %s" % err)

        self._vmpool.add_vm(vm)
        gc3libs.log.info(
            "VM with id `%s` has been created and is in %s state.",
            vm.id, vm.status)
        return vm

    def _import_keypair(self):
        """
        Create a new keypair and import the public key defined in the
        configuration file.
        """
        fd = open(os.path.expanduser(self.public_key))
        try:
            key_material = fd.read()
            self.compute_client.keypairs.create(self.keypair_name, key_material)
            keypair = self.compute_client.keypairs.get(self.keypair_name)
            gc3libs.log.info(
                "Successfully imported key `%s` with fingerprint `%s`"
                " as keypair `%s`" % (self.public_key,
                                      keypair.fingerprint,
                                      self.keypair_name))
            return keypair
        except Exception as ex:
            fd.close()
            raise UnrecoverableError("Error importing keypair %s: %s"
                                     % (self.keypair_name, ex))

    def _get_subresource(self, vm):
        """
        Return the resource associated to the virtual machine `vm`.

        Updates the internal list of available resources if needed.
        """
        self._connect()
        if vm.id not in self.subresources:
            self.subresources[vm.id] = self._make_subresource(
                vm.id, vm.preferred_ip)
            # Update resource based on flavor specs
            try:
                flavor = self.compute_client.flavors.get(vm.flavor['id'])
                res = self.subresources[vm.id]
                res['max_memory_per_core'] = flavor.ram * MiB
                res['max_cores_per_job'] = flavor.vcpus
                res['max_cores'] = flavor.vcpus
            except Exception as ex:
                # Ignore any error here, as we can get information on
                # the subresources when we connect to it.
                gc3libs.log.info(
                    "Ignoring error in setting max_cores_per_job/max_memory"
                    " values for new subresource %s based on flavor: %s",
                    self.subresources[vm.id].name, ex)
        return self.subresources[vm.id]

    def _get_vm(self, vm_id):
        """
        Return the instance with id `vm_id`, raises an error if there
        is no such instance with that id.
        """
        self._connect()
        vm = self._vmpool.get_vm(vm_id)
        return vm

    @staticmethod
    def __str_fingerprint(pkey):
        """
        Print key fingerprint like SSH commands do.

        We need to convert the key fingerprint from Paramiko's
        internal representation to this colon-separated hex format
        because that's the way the fingerprint is returned from the
        OpenStack API.
        """
        return ':'.join((i.encode('hex') for i in pkey.get_fingerprint()))

    def _have_keypair(self, keypair):
        """
        Check if the given SSH key is available locally.

        Try to locate the given SSH key first among the keys loaded
        into a running ``ssh-agent`` (if any), then in the key file
        given in the ``public_key`` configuration key.
        """
        # try with SSH agent first
        gc3libs.log.debug("Checking if keypair is registered in SSH agent...")
        agent = paramiko.Agent()
        fingerprints_from_agent = [self.__str_fingerprint(k)
                                   for k in agent.get_keys()]
        if keypair.fingerprint in fingerprints_from_agent:
            gc3libs.log.debug("Found remote key fingerprint in SSH agent.")
            return True

        # else, try to load from file
        keyfile = self.public_key
        if keyfile.endswith('.pub'):
            keyfile = keyfile[:-4]
        else:
            gc3libs.log.warning(
                "Option `public_key` in configuration file should contain"
                " the path to a public key file (with `.pub` ending),"
                " but `%s` was found instead. Continuing anyway.",
                self.public_key)

        pkey = None
        for format, reader in [
                ('DSS', paramiko.DSSKey.from_private_key_file),
                ('RSA', paramiko.RSAKey.from_private_key_file),
        ]:
            try:
                gc3libs.log.debug(
                    "Trying to load key file `%s` as SSH %s key...",
                    keyfile, format)
                pkey = reader(keyfile)
                gc3libs.log.info(
                    "Successfully loaded key file `%s` as SSH %s key.",
                    keyfile, format)
            # PasswordRequiredException < SSHException so we must
            # check this first
            except paramiko.PasswordRequiredException:
                gc3libs.log.warning(
                    "Key %s is encripted with a password, so we cannot check"
                    " if it matches the remote keypair (maybe you should start"
                    " `ssh-agent`?). Continuing without consistency check with"
                    " remote fingerprint.", keyfile)
                return False
            except paramiko.SSHException as ex:
                gc3libs.log.debug(
                    "File `%s` is not a valid %s private key: %s",
                    keyfile, format, ex)
                # try with next format
                continue
        if pkey is None:
            raise ValueError("Public key `%s` is neither a valid"
                             " RSA key nor a DSS key" % self.public_key)

        # check key fingerprint
        localkey_fingerprint = self.__str_fingerprint(pkey)
        if localkey_fingerprint != keypair.fingerprint:
            raise UnrecoverableAuthError(
                "Keypair `%s` exists but has different fingerprint than local"
                " one: local public key file `%s` has fingerprint `%s`,"
                " whereas OpenStack API reports fingerprint `%s`."
                " Aborting!" % (
                    self.public_key,
                    self.keypair_name,
                    localkey_fingerprint,
                    keypair.fingerprint,
                ),
                do_log=True)

    @cache_for(120)
    def _get_security_groups(self):
        self._connect()
        try:
            # python-novaclient < 8.0.0
            return self.compute_client.security_groups.list()
        except AttributeError:
            return self.network_client.list_security_groups()['security_groups']

    @cache_for(120)
    def _get_security_group(self, name):
        groups = self._get_security_groups()
        group = [grp for grp in groups if lookup(grp, 'name') == name]
        if not group:
            raise NotFound("Security group %s not found." % name)
        return group[0]

    def _make_subresource(self, id, remote_ip):
        """
        Create a resource associated to the instance with `remote_ip`
        ip using configuration file parameters.
        """
        gc3libs.log.debug(
            "Creating remote ShellcmdLrms resource for IP address %s",
            remote_ip)
        args = self.subresource_args.copy()
        args['frontend'] = remote_ip
        args['name'] = "%s@%s" % (id, self.name)
        return ShellcmdLrms(**args)

    def _parse_security_group(self):
        """
        Parse configuration file and set `self.security_group_rules`
        with a list of dictionaries containing the rule sets
        """
        rules = self.security_group_rules.split(',')
        self.security_group_rules = []
        for rule in rules:
            rulesplit = rule.strip().split(':')
            if len(rulesplit) != 4:
                gc3libs.log.warning("Invalid rule specification in"
                                    " `security_group_rules`: %s" % rule)
                continue
            self.security_group_rules.append({
                'ip_protocol': rulesplit[0],
                'from_port': int(rulesplit[1]),
                'to_port': int(rulesplit[2]),
                'cidr_ip': rulesplit[3],
            })

    def _setup_security_groups(self):
        """
        Check the current configuration and set up the security group
        if it does not exist.
        """
        self._connect()
        if not self.security_group_name:
            gc3libs.log.error("Group name in `security_group_name`"
                              " configuration option cannot be empty!")
            return

        try:
            self._get_security_group(self.security_group_name)
        except NotFound:
            try:
                gc3libs.log.info("Creating security group %s",
                                 self.security_group_name)

                self.compute_client.security_groups.create(
                    self.security_group_name,
                    "GC3Pie_%s" % self.security_group_name)
            except Exception as ex:
                gc3libs.log.error("Error creating security group %s: %s",
                                  self.security_group_name, ex)
                raise UnrecoverableError(
                    "Error creating security group %s: %s"
                    % (self.security_group_name, ex))

            self._get_security_group(self.security_group_name)
        # TODO: Check if the security group has all the rules we want
        # security_group = groups[self.security_group_name]
        # current_rules = []
        # for rule in security_group.rules:
        #     rule_dict = {
        #         'ip_protocol':  rule.ip_protocol,
        #         'from_port':    int(rule.from_port),
        #         'to_port':      int(rule.to_port),
        #         'cidr_ip':      str(rule.grants[0]),
        #     }
        #     current_rules.append(rule_dict)

        # for new_rule in self.security_group_rules:
        #     if new_rule not in current_rules:
        #         security_group.authorize(**new_rule)

    @cache_for(120)
    def _get_available_images(self):
        self._connect()
        try:
            # python-novaclient < 8.0.0
            return self.compute_client.images.list()
        except AttributeError:
            # ``glance_client.images.list()`` returns a generator, but callers
            # of `._get_available_images()` expect a Python list
            return list(self.image_client.images.list())

    @cache_for(120)
    def _get_available_flavors(self):
        self._connect()
        return self.compute_client.flavors.list()

    def _get_flavor(self, name):
        for flavor in self._flavors:
            # pick the first match by that name
            if flavor.name == name:
                return flavor
        # no flavor by the given name
        raise NotFound("Flavor `{name}` not found.".format(name=name))

    @cache_for(120)
    def _get_keypair(self, keypair_name):
        self._connect()
        return self.compute_client.keypairs.get(keypair_name)

    def get_image_id_for_job(self, job):
        """
        If a configuration option <application>_image_id is present,
        returns its value, otherwise returns `self.image_id`
        """
        conf_option = job.application_name + '_image_id'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.image_id

    def get_instance_type_for_job(self, task):
        """
        Return the "best" instance_type (aka flavor) for the task.

        This method will always returns an instance_type that accomodates the
        task's requirements. Preference will be given, in order, to:

        1) ``<application>_instance_type configuration`` option
        2) ``instance_type configuration`` option
        3) the smallest flavor (ordering by cpus, ram and disk)
           that fits the application requirements.
        """
        req_cores = self._get_task_requirement(task, 'requested_cores', 1)
        req_memory = self._get_task_requirement(task, 'requested_memory', 0 * MiB)

        # If there is an option <application>_instance_type, try to use it
        for conf_option in [
                task.application_name + '_instance_type',
                'instance_type'
        ]:
            gc3libs.log.debug("Trying configuration option `%s` ...", conf_option)
            try:
                flavor_name = self[conf_option]
            except KeyError:
                # ignore and try next option
                continue
            if flavor_name is None:
                continue

            try:
                flavor = self._get_flavor(flavor_name)
            except NotFound:
                gc3libs.log.warning(
                    "Unable to use flavor `%s` for task `%s` as requested"
                    " by config option `%s` because it's not available"
                    " in the cloud backend. Fix configuration file?",
                    flavor_name, task, conf_option)
                continue

            if self._flavor_matches_requirements(flavor, req_cores, req_memory):
                gc3libs.log.info(
                    "Using flavor `%s` for task `%s`"
                    " as requested by config option `%s`",
                    flavor.name, task, conf_option)
                return flavor
            else:
                gc3libs.log.warning(
                    "Unable to use flavor `%s` for task `%s`"
                    " as requested by config option `%s`"
                    " because it doesn't fit application's requirements",
                    flavor_name, task, conf_option)
                continue

        # fall back to smallest flavor that fits
        valid_flavors = [
            flv for flv in self._flavors
            if self._flavor_matches_requirements(flv, req_cores, req_memory)
        ]
        flavor = min(valid_flavors, key=self._flavor_ordering_key)
        gc3libs.log.info(
            "Using VM flavor `%s` as the smallest flavor that fits requirements"
            " of task `%s`", flavor.name, task)
        return flavor

    @staticmethod
    def _get_task_requirement(task, reqname, default):
        """
        Return attribute `reqname` of `task`.

        If the attribute is not set or is ``None``, then return `default`
        instead.
        """
        req = getattr(task, reqname, None)
        if req is None:
            req = default
        return req

    def _flavor_matches_requirements(self, flavor, req_cores, req_memory):
        """
        Return whether the flavor matches the given requirements.
        """
        return (flavor.vcpus >= req_cores and
                (flavor.ram * MiB - self.vm_os_overhead) >= req_memory)

    @staticmethod
    def _flavor_ordering_key(flavor):
        """
        Return the value to use when sorting flavors.

        By default, flavors are sorted lexicographically by:
        - nr. of vCPUS (first),
        - amount of RAM,
        - size of disk (last).
        """
        return (flavor.vcpus, flavor.ram, flavor.disk)

    def get_user_data_for_job(self, job):
        """
        If a configuration option <application>_user_data is present,
        returns its value, otherwise returns `self.user_data`.
        """
        conf_option = job.application_name + '_user_data'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.user_data

    @same_docstring_as(LRMS.cancel_job)
    def cancel_job(self, app):
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
            return subresource.cancel_job(app)
        except InstanceNotFound:
            # ignore -- if this VM exists no more, we need not cancel any job
            pass
        except UnrecoverableError as err:
            gc3libs.log.error(
                "Changing state of task '%s' to UNKNOWN because of"
                " an OpenStack API error (%s: %s)",
                app.execution.lrms_jobid, err.__class__.__name__, err)
            app.execution.state = Run.State.UNKNOWN
            app.execution.history.append(
                "State changed to UNKNOWN because of"
                " an OpenStack API error (%s: %s)."
                % (err.__class__.__name__, err))
            raise

    @same_docstring_as(LRMS.get_resource_status)
    def get_resource_status(self):
        self.updated = False
        # Since we create the resource *before* the VM is actually up
        # & running, it's possible that the `frontend` value of the
        # resources points to a non-existent hostname. Therefore, we
        # have to update them with valid public_ip, if they are
        # present.

        # Update status of known VMs
        for vm_id in self._vmpool:
            try:
                vm = self._vmpool.get_vm(vm_id, force_reload=True)
            except UnrecoverableError:
                gc3libs.log.warning(
                    "Removing stale information on VM `%s`. It has probably"
                    " been deleted from outside GC3Pie.", vm_id)
                self._vmpool.remove_vm(vm_id)
                continue

            if vm.status in PENDING_STATES:
                # If VM is still in pending state, skip creation of
                # the resource
                continue
            elif vm.status in ERROR_STATES:
                # The VM is in error state: exit.
                gc3libs.log.error(
                    "VM with id `%s` is in ERROR state."
                    " Terminating it!", vm.id)
                vm.delete()
                self._vmpool.remove_vm(vm.id)
                self.subresources.pop(vm.id)
                continue
            elif vm.status == 'DELETED':
                gc3libs.log.info(
                    "VM `%s` in DELETE state. It has probably been terminated"
                    " from outside GC3Pie. Removing it from the list of VM.",
                    vm.id)
                self._vmpool.remove_vm(vm.id)
                self.subresources.pop(vm.id)
                continue
            elif vm.status in ['SHUTOFF', 'SUSPENDED',
                               'RESCUE', 'VERIFY_RESIZE']:
                # The VM has probably ben stopped or shut down from
                # outside GC3Pie.
                gc3libs.log.error(
                    "VM with id `%s` is in permanent state `%s`.",
                    vm.id, vm.status)
                continue

            # Get or create a resource associated to the vm
            subresource = self._get_subresource(vm)
            try:
                subresource.get_resource_status()
            except TransportError as ex:
                # TODO: get all the IPs and try with all of them to connect.
                # Start with preferred_ip if defined
                gc3libs.log.info(
                    "Ignoring error in updating resource '%s': %s."
                    " Trying other IPs.", subresource.name, ex)
                for ip in sum(list(vm.networks.values()), []):
                    if vm.preferred_ip == ip:
                        continue
                    vm.preferred_ip = ip
                    subresource.frontend = ip
                    gc3libs.log.info(
                        "Connection error. Trying with alternate IP address "
                        "%s", vm.preferred_ip)
                    try:
                        subresource.get_resource_status()
                        break
                    except Exception as ex:
                        gc3libs.log.info(
                            "Ignoring error in updating resource '%s': %s."
                            " The corresponding VM may not be ready yet.",
                            subresource.name, ex)
                # Unable to connect to the VM using any IP.  Ensure
                # this resource is considered "pending" as we couldn't
                # update its status
                subresource.updated = False
            except Exception as ex:
                # XXX: Actually, we should try to identify the kind of
                # error we are getting. For instance, if the
                # configuration options `username` is wrong, we will
                # create VMs but we will never be able to submit jobs
                # to them, thus causing an increasing number of
                # useless VMs created on the cloud.
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'openstack',
                        # - class
                        'OpenStackLrms',
                        # - method
                        'get_resource_status',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'resource',
                        'status',
                        'update',
                        'vm',
                ):
                    gc3libs.log.info(
                        "Ignoring error while updating resource '%s'. "
                        "The corresponding VM may not be ready yet. Error: %s",
                        subresource.name, ex)
                else:
                    # propagate exception back to caller
                    raise
        self._vmpool.update()
        return self

    @same_docstring_as(LRMS.get_results)
    def get_results(self, app, download_dir, overwrite=False,
                    changed_only=True):
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
        except InstanceNotFound:
            gc3libs.log.error(
                "Changing state of task '%s' to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore.",
                app.execution.lrms_jobid, app.execution._lrms_vm_id)
            app.execution.state = Run.State.TERMINATED
            app.execution.signal = Run.Signals.RemoteError
            app.execution.history.append(
                "State changed to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore."
                % (app.execution._lrms_vm_id,))
            raise UnrecoverableDataStagingError(
                "VM where job was running is no longer available")
        except UnrecoverableError as err:
            gc3libs.log.error(
                "Changing state of task '%s' to UNKNOWN because of"
                " an OpenStack API error (%s: %s)",
                app.execution.lrms_jobid, err.__class__.__name__, err)
            app.execution.state = Run.State.UNKNOWN
            app.execution.history.append(
                "State changed to UNKNOWN because of"
                " an OpenStack API error (%s: %s)."
                % (err.__class__.__name__, err))
            raise
        return subresource.get_results(
            app, download_dir, overwrite, changed_only)

    @same_docstring_as(LRMS.update_job_state)
    def update_job_state(self, app):
        self._connect()
        if app.execution._lrms_vm_id not in self.subresources:
            try:
                self.subresources[app.execution._lrms_vm_id] = self._get_subresource(
                    self._get_vm(app.execution._lrms_vm_id))
            except InstanceNotFound:
                gc3libs.log.error(
                    "Changing state of task '%s' to TERMINATED since OpenStack"
                    " instance '%s' does not exist anymore.",
                    app.execution.lrms_jobid, app.execution._lrms_vm_id)
                app.execution.state = Run.State.TERMINATED
                app.execution.signal = Run.Signals.RemoteError
                app.execution.history.append(
                    "State changed to TERMINATED since OpenStack"
                    " instance '%s' does not exist anymore."
                    % (app.execution._lrms_vm_id,))
                raise
            except UnrecoverableError as err:
                gc3libs.log.error(
                    "Changing state of task '%s' to UNKNOWN because of"
                    " an OpenStack API error (%s: %s)",
                    app.execution.lrms_jobid, err.__class__.__name__, err)
                app.execution.state = Run.State.UNKNOWN
                app.execution.history.append(
                    "State changed to UNKNOWN because of"
                    " an OpenStack API error (%s: %s)."
                    % (err.__class__.__name__, err))
                raise
        return self.subresources[app.execution._lrms_vm_id].update_job_state(app)

    def submit_job(self, job):
        """
        Submission on an OpenStack resource will usually happen in
        multiple steps, since creating a VM and attaching a resource
        to it will take some time.

        In order to return as soon as possible, the backend will raise
        a `RecoverableError` whenever submission is delayed.

        In case a permanent error is found (for instance, we cannot
        create VMs on the cloud), a `UnrecoverableError` is raised.

        More in detail, when during submission the following will
        happen:

        * First of all, the backend will try to submit the job to one
          of the already available subresources.

        * If none of them is able to sbmit the job, the backend will
          check if there is a VM in pending state, and in case there
          is one it will raise a `RecoverableError`, thus delaying
          submission.

        * If no VM in pending state is found, the `vm_pool_max_size`
          configuration option is checked. If we already reached the
          maximum number of VM, a `UnrecoverableError` is raised.

        * If no VM in pending state is found but `vm_pool_max_size` is
          still lesser than the number of VM currently created (or it
          is None, which for us means no limit), then a new VM is
          created, and `RecoverableError` is raised.

        """
        # Updating resource is needed to update the subresources. This
        # is not always done before the submit_job because of issue
        # nr.  386:
        #     https://github.com/uzh/gc3pie/issues/386
        self.get_resource_status()
        pending_vms = set(vm.id for vm in self._vmpool.get_all_vms()
                          if vm.status in PENDING_STATES)

        image_id = self.get_image_id_for_job(job)
        # Check if the image id is valid
        if image_id not in [img.id for img in self._get_available_images()]:
            raise ConfigurationError("Image ID %s not found in cloud "
                                     "%s" % (image_id, self.os_auth_url))

        instance_type = self.get_instance_type_for_job(job)
        if not instance_type:
            raise RuntimeError(
                "Unable to find a suitable instance type for "
                "application %s" % job)

        # First of all, try to submit to one of the subresources.
        for vm_id, subresource in list(self.subresources.items()):
            if not subresource.updated:
                # The VM is probably still booting, let's skip to the
                # next one and add it to the list of "pending" VMs.
                pending_vms.add(vm_id)
                continue
            try:
                # Check that the required image id and instance type
                # are correct
                vm = self._get_vm(vm_id)
                if vm.image['id'] != image_id:
                    continue
                subresource.submit_job(job)
                job.execution._lrms_vm_id = vm_id
                job.changed = True
                gc3libs.log.info(
                    "Job successfully submitted to remote resource %s.",
                    subresource.name)
                return job
            except (LRMSSubmitError, InstanceNotFound) as ex:
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'openstack',
                        # - class
                        'OpenStackLrms',
                        # - method
                        'submit_job',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'submit',
                ):
                    gc3libs.log.debug(
                        "Ignoring error in submitting to resource '%s': %s",
                        subresource.name, ex)
                else:
                    # propagate error back to caller
                    raise

        # Couldn't submit to any resource.
        if not pending_vms:
            # No pending VM, and no resource available. Create a new VM
            if not self.vm_pool_max_size \
                    or len(self._vmpool) < self.vm_pool_max_size:
                user_data = self.get_user_data_for_job(job)
                vm = self._create_instance(
                    image_id,
                    name="GC3Pie_%s_%d" % (self.name, (len(self._vmpool) + 1)),
                    instance_type=instance_type,
                    user_data=user_data)
                pending_vms.add(vm.id)

                self._vmpool.add_vm(vm)
            else:
                raise MaximumCapacityReached(
                    "Already running the maximum number of VM on resource %s:"
                    " %d VMs started, but max %d allowed by configuration."
                    % (self.name, len(self._vmpool), self.vm_pool_max_size),
                    do_log=True)

        # If we reached this point, we are waiting for a VM to be
        # ready, so delay the submission until we wither can submit to
        # one of the available resources or until all the VMs are
        # ready.
        gc3libs.log.debug(
            "No available resource was found, but some VM is still in"
            " `pending` state. Waiting until the next iteration before"
            " creating a new VM. Pending VM ids: %s", pending_vms)
        raise ResourceNotReady(
            "Delaying submission until one of the VMs currently pending"
            " is ready. (Pending VM ids: %r)" % (pending_vms,))

    @same_docstring_as(LRMS.peek)
    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
        except InstanceNotFound:
            gc3libs.log.error(
                "Changing state of task '%s' to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore.",
                app.execution.lrms_jobid, app.execution._lrms_vm_id)
            app.execution.state = Run.State.TERMINATED
            app.execution.signal = Run.Signals.RemoteError
            app.execution.history.append(
                "State changed to TERMINATED since OpenStack"
                " instance '%s' does not exist anymore."
                % (app.execution._lrms_vm_id,))
            raise UnrecoverableDataStagingError(
                "VM where job was running is no longer available.")
        return subresource.peek(app, remote_filename, local_file, offset, size)

    def validate_data(self, data_file_list=None):
        """
        Supported protocols: file
        """
        for url in data_file_list:
            if url.scheme not in ['file', 'http', 'https', 'swift', 'swifts', 'swt', 'swts']:
                return False
        return True

    def free(self, app):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should terminate any remote instance used
        by the job.

        Call this method when `app.execution.state` is anything other
        than `TERMINATED` results in undefined behavior and will
        likely be the cause of errors later on.  Be cautious.
        """
        # XXX: this should probably done only when no other VMs are
        # using this resource.

        # FIXME: freeing the resource from the application is probably
        # not needed since instances are not persistent.

        # freeing the resource from the application is now needed as
        # the same instanc may run multiple applications
        try:
            subresource = self._get_subresource(
                self._get_vm(app.execution._lrms_vm_id))
        except InstanceNotFound:
            # ignore -- if the instance is no more, there is
            # nothing we should free
            return
        subresource.free(app)

        # FIXME: current approach in terminating running instances:
        # if no more applications are currently running, turn the instance off
        # check with the associated resource
        subresource.get_resource_status()
        if not subresource.has_running_tasks():
            # turn VM off
            vm = self._get_vm(app.execution._lrms_vm_id)

            gc3libs.log.info("VM instance %s at %s is no longer needed."
                             " Terminating.", vm.id, vm.preferred_ip)
            del self.subresources[vm.id]
            vm.delete()
            del self._vmpool[vm.id]

    @same_docstring_as(LRMS.close)
    def close(self):
        gc3libs.log.info("Closing connection to cloud '%s'...",
                         self.name)
        if not self.enabled:
            # The resources was most probably disabled by command
            # line. We didn't update it before, so we don't care about
            # currently running VMs now.
            return
        # Update status of VMs and remote resources
        self.get_resource_status()
        for vm_id, subresource in list(self.subresources.items()):
            if subresource.updated and not subresource.has_running_tasks():
                vm = self._get_vm(vm_id)
                gc3libs.log.warning(
                    "VM instance %s at %s is no longer needed. "
                    "You may need to terminate it manually.",
                    vm.id, vm.preferred_ip)
                self._vmpool.remove_vm(vm.id)
            subresource.close()

    def __getstate__(self):
        # Do not save `novaclient.client.Client` class as it is
        # not pickle-compliant
        state = self.__dict__.copy()
        del state['client']
        del state['__connected']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.compute_client, self.image_client, self.network_client = self._new_clients()
        self._connect()


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="openstack",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   grosetta.py -- Front-end script for submitting ROSETTA jobs to SMSCG.
#
#   Copyright (C) 2010, 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting ROSETTA jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``grosetta --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2012-02-28:
    * Use option ``-R`` to request a specific release of Rosetta.
  2011-03-28:
    * No longer require any input file on the command line: e.g., ``grosetta -l``
      will just list the status of jobs in the current session.  On the other hand,
      if the FLAGS file is given, then given must be also at least one input file.
  2010-12-20:
    * Initial release, forking off the old `grosetta`/`gdocking` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'



# workaround Issue 95, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import grosetta2015
    grosetta2015.GRosettaScript().run()


# stdlib imports
import os
import os.path
import sys

from pkg_resources import Requirement, resource_filename

# interface to Gc3libs
import gc3libs
from gc3libs.application.rosetta import RosettaApplication
from gc3libs.cmdline import SessionBasedScript, positive_int
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

INPUT_PDB_EXTENSION="pdb"
ROSETTA_OPTION_FILE="options"
INPUT_LIST_PATTERNS = [INPUT_PDB_EXTENSION,ROSETTA_OPTION_FILE]

## the application class
## custom application class
class Rosetta2015Application(Application):
    """
    """
    application_name = 'grosetta2015'

    def __init__(self, input_folder, command_to_run, **extra_args):

        inputs = dict()
        outputs = dict()

        # grosetta2015_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
        #                                       "gc3libs/etc/rosetta2015.sh")
        # inputs[grosetta2015_wrapper_sh] = os.path.basename(grosetta2015_wrapper_sh)

        for input in os.listdir(input_folder):
            inputs[os.path.join(input_folder,input)] = os.path.basename(input)

        # arguments = "./%s --no-tar @options" % (inputs[grosetta2015_wrapper_sh])

        Application.__init__(
            self,
            arguments = command_to_run,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'grosetta2015.log',
            join=True,
            **extra_args)

## the script class
class GRosettaScript(SessionBasedScript):
    """
Run MiniRosetta on the specified INPUT files and fetch OUTPUT files
back from the execution machine; if OUTPUT is omitted, all '*.pdb',
'*.sc' and '*.fasc' files are retrieved.  Several instances can be run in
parallel, depending on the '-P' and '-p' options.

The `grosetta` command keeps a record of jobs (submitted, executed and
pending) in a session file (set name with the '-s' option); at each
invocation of the command, the status of all recorded jobs is updated,
output from finished jobs is collected, and a summary table of all
known jobs is printed.  New jobs are added to the session if the number
of wanted decoys (option '-P') is raised.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; `grosetta` will delay submission
of newly-created jobs so that this limit is never exceeded.

Note: the list of INPUT and OUTPUT files must be separated by ':'
(on the shell command line, put a space before and after).
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            usage = "%(prog)s [options] FLAGSFILE INPUT ... [: OUTPUT ...]",
            version = __version__, # module version == script version
            application = RosettaApplication,
            )

    def setup_args(self):

        self.add_param('input_pdb', type=str,
                       help="Root localtion of input .pdb pais.")

        self.add_param('command_to_run', type=str,
                       help="Rosetta related command to be executed. "
                       "Note: no control over the consistency nor the availability "
                       " of the command will be made.")

    def parse_args(self):

        self.input_pdbs = dict()

        self.input_folders = [ os.path.join(self.params.input_pdb,folder) for folder in os.listdir(self.params.input_pdb) ]

        # for r,d,f in os.walk(self.params.input_pdb):
        #     input_pdb = []
        #     for infile in f:
        #         if any(infile.endswith(pattern) for pattern in INPUT_LIST_PATTERNS):
        #             input_pdb.append(os.path.join(r,infile))
        #     # XXX: Validation of input files to be done
        #     if input_pdb:
        #         self.input_pdbs[r] = input_pdb

        self.log.debug("Found '%d' valid input folders", len(self.input_folders))

    def new_tasks(self, extra):
        tasks = []

        for input_folder in self.input_folders:
            # extract root folder name to be used as jobname
            pairname = input_folder

            extra_args = extra.copy()
            extra_args['jobname'] = pairname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % pairname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % pairname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % pairname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % pairname)

            gc3libs.log.debug("Adding task for folder %s" % input_folder)

            tasks.append(Rosetta2015Application(
                input_folder,
                self.params.command_to_run,
                **extra_args))
        return tasks



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'

from gc3libs.cmdline import SessionBasedScript
from gc3libs import Application


class SimpleScript(SessionBasedScript):

    """stupid class"""
    version = '1'

    def new_tasks(self, extra):

        default_output_dir = extra.pop('output_dir')
        return [
            # old-style
            ('MyJob',
             Application,
               [
                   # arguments
                   ('/bin/bash', '-c', 'echo ciao > SimpleScript.stdout'),
                   # inputs
                   [],
                   # outputs
                   ['SimpleScript.stdout'],
               ],
             dict(output_dir='SimpleScript.out.d', join=True)),
            # new style, with explicit output dir
            Application(
                # arguments
                ('/bin/bash', '-c', 'echo ciao > SimpleScript.stdout'),
                # inputs
                [],
                # outputs
                ['SimpleScript.stdout'],
                # output_dir
                'SimpleScript.out2.d',
                # extra args
                join=True,
                **extra
            ),
            # new style, with output dir gotten from `self.extra`
            Application(
                # arguments
                ('/bin/bash', '-c', 'echo ciao > SimpleScript.stdout'),
                # inputs
                [],
                # outputs
                ['SimpleScript.stdout'],
                join=True,
                output_dir=default_output_dir,
                **extra
            ),
        ]

# main: run tests

if "__main__" == __name__:
    app = SimpleScript()
    app.run()
